"use strict";

//星クラス
class Star {
    constructor() {
        this.x = Math.random() * field.width;//X座標
        this.y = Math.random() * field.height;//Y座標
        this.r = Math.random() * 5 + 1;//半径
    }
    //フレーム(画面)ごとに呼び出されるメソッド
    tick() {
        this.y += this.r;//星を半径分下に移動
        //画面下部に来たら上に移動
        if (this.y > field.height) {
            this.y -= field.height;
        }
        drawCircle(this.x, this.y, this.r, "#888800");
    }
}

//自機クラス
class Ship {
    constructor() {
        this.img = document.getElementById("ship");
        this.x = field.width / 2;
        this.y = field.height / 1.5;
        this.sx = 0;
        this.sy = 0;
    }
    move(mouseX, mouseY) {
        this.sx = (mouseX - this.x) / 10;//マウスのX軸方向に向かって移動
        this.sy = (mouseY - this.y) / 10;//マウスのY軸方向に向かって移動
    }
    //フレーム(画面)ごとに呼び出されるメソッド
    tick() {
        this.x += this.sx;//速度sxを座標xに反映
        this.y += this.sy;//速度syを座標yに反映
        this.x = this.x < 100 ? 100 : this.x;
        this.y = this.y < 0 ? 0 : this.y;
        ctx.drawImage(this.img, this.x - 100, this.y);
    }
    //弾丸を発射
    shoot() {
        bullets.push(new Bullet((this.x) - 50, this.y, 0, -25, true));
    }
}

//敵クラス
class Enemy {
    constructor() {
        this.img = document.getElementById("enemy");
        this.x = Math.random() * field.width / 1.2 + 100;//X座標
        this.y = 0;//Y座標
        this.sx = Math.random() * 5 - 2.5;//X軸方向初速
        this.sy = Math.random() * 15 + 15;//Y軸方向初速
        this.shoot = false;
    }
    //フレーム(画面)ごとに呼び出されるメソッド
    tick() {
        //X方向は一定の速度、Y方向は速度を減らしながら移動することで放物線の様な軌跡を描く
        this.sy -= 1;//速度を減らす
        this.x += this.sx;//速度sxを座標xに反映
        this.y += this.sy;//速度syを座標yに反映
        ctx.drawImage(this.img, this.x - 50, this.y - 50);

        if (this.shoot == false && this.sy < 0) {
            //速度が上向きになったタイミングで自機に向かって弾丸を発射
            let theta = Math.atan2(ship.y - this.y, ship.x - this.x);
            let sx = Math.cos(theta) * 10;
            let sy = Math.sin(theta) * 10;
            bullets.push(new Bullet(this.x, this.y, sx, sy, false));
            this.shoot = true;
        }
    }
}

//弾丸クラス
class Bullet {
    constructor(x, y, sx, sy, isShip) {
        this.x = x;
        this.y = y;
        this.sx = sx;
        this.sy = sy;
        this.isShip = isShip;//自機の弾丸か敵の弾丸かを区別するためのプロパティ
    }
    tick() {
        this.x += this.sx;
        this.y += this.sy;
        drawCircle(this.x, this.y, 5, this.isShip ? "yellow" : "red");
    }
}

//(x,y)を中心に半径r、色colorの円を描画
function drawCircle(x, y, r, color) {
    ctx.fillStyle = color;
    ctx.beginPath();
    ctx.moveTo(x, y);
    ctx.arc(x, y, r, 0, Math.PI * 2);
    ctx.closePath();
    ctx.fill();
}

let ctx;//描画コンテキスト
let ship;//自機
let back;//背景画像
let count = 0;//敵出現用カウンタ
let interval = 50;//敵出現頻度
let timerId;//タイマー
let bullets = [];//弾丸のリスト
let enemies = [];//敵のリスト
const stars = [];//星のリスト
let score = 0;//スコア
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let field = document.getElementById("field");//<canvas>要素への参照を取得

onload = () => {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    field.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    field.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = field.getContext("2d");//描画コンテキスト
    ctx.font = "2rem 'Times New Roman'";
    ship = new Ship();//自機オブジェクト作成
    back = document.getElementById("back");
    //イベントハンドラの登録
    //マウス移動⇒自機を移動
    window.onpointermove = (e) => {
        ship.move(e.clientX, e.clientY);
    };
    //マウス押下⇒弾丸発射
    window.onpointerdown = (e) => {
        ship.shoot();
    };
    //Enterキー押下⇒弾丸発射
    window.onkeydown = (e) => {
        if (e.code == "Enter") {
            ship.shoot();
        }
    };
    timerId = setInterval(tick, 50);//タイマー開始
    for (let i = 0; i < 50; i++) {
        stars.push(new Star());//星のオブジェクトを作成して配列starsに追加
    }
}

function tick() {
    count++;
    ctx.fillStyle = "black";
    ctx.fillRect(0, 0, field.width, field.height);
    ctx.drawImage(back, 0, 0, field.width, field.height);
    stars.forEach((s) => s.tick());//星の移動と描画
    ship.tick();
    if (count % interval == 0) {
        enemies.push(new Enemy());//intervalフレーム毎に敵オブジェクトを作成して配列enemiesに追加
        interval = Math.max(5, interval - 5);
    }
    let gameOver = false;
    enemies.forEach((e) => {
        e.tick();//敵を移動
        //自機と敵との距離が100未満⇒ゲームオーバー
        if (dist(e, ship) < 100) {
            gameOver = true;
        }
    });
    bullets.forEach((b) => {
        b.tick();//弾丸移動
        //敵の弾丸と自機との距離が30未満⇒ゲームオーバー
        if (!b.isShip && dist(b, ship) < 30) {
            gameOver = true;
        }
    });
    let prevNum = enemies.length;//撃墜される前の敵の配列enemiesの要素の数
    //配列enemiesのfilter()メソッドを使って、自機の弾丸と衝突していない敵のみが残るようにし、配列bulletsから個々の弾丸を取り出し、some()メソッドを使って、自機の発射した弾丸との距離が50未満の場合は、敵を撃ち落としたとして処理
    //some()メソッド:配列内のいずれかの要素が条件に合致する場合にtrueを返し、全ての要素が条件に合致しない場合にのみfalseを返す
    enemies = enemies.filter((e) => {
        return !bullets.some((b) => {
            return b.isShip && dist(e, b) < 50;//自機の弾丸と敵の衝突判定
        });
    });
    score += prevNum - enemies.length;//撃墜前と撃墜後の敵の配列enemiesの要素の数の比較
    ctx.fillStyle = "yellow";
    ctx.fillText("スコア:　" + score, field.width / 2.4, 50);

    if (gameOver) {
        clearInterval(timerId);
        ctx.fillStyle = "yellow";
        ctx.fillText("ゲームオーバー", field.width / 2.5, field.height / 2);
    }
}

//2つのオブジェクト間の距離を求める
//Math.sqrt()関数は、引数の平方根を返す
function dist(e0, e1) {
    return Math.sqrt(
        Math.abs(e0.x - e1.x) ** 2 + Math.abs(e0.y - e1.y) ** 2
    );
}
