"use strict";

let scroller = new Scroller();//スクロール処理するオブジェクト
Player.prototype = scroller;//scrollerをPlayerのprototypeに設定
Alien.prototype = scroller;//scrollerをAlienのprototypeに設定

const W = 31;//迷路の幅
const H = 31;//迷路の高さ
const GAMECLEAR = 1;//ゲームクリア状態
const GAMEOVER = 2;//ゲームオーバー状態
const maze = [];//迷路を保持する２次元配列
const player = new Player(1, 1);//主人公オブジェクト
const aliens = [new Alien(W - 2, 1), new Alien(1, W - 2)];//敵オブジェクトを含む配列

let ctx;//描画用コンテキスト
let code = 0;//現在どのキーが押されているか
let game = 0;//ゲームの状態。クリア＝１、ゲームオーバー＝２
let timer = NaN;//タイマー

//スクロール処理オブジェクト
function Scroller() {
    //doScroll()メソッドは、スクロール用カウンタscrollCountを増加させる
    //移動先(dx,dy)が設定された場合にスクロール
    this.doScroll = function () {
        if (this.dx == 0 && this.dy == 0) {
            return;//移動先(dx,dy)が共に０のときは何もしない
        }

        //スクロール用カウンタscrollCountの値が５以上になった場合に一コマ移動させる
        if (++this.scrollCount >= 5) {
            this.x = this.x + this.dx;//X座標更新
            this.y = this.y + this.dy;//Y座標更新
            //移動先(dx,dy)とスクロール用カウンタscrollCountを０で初期化
            this.dx = 0;
            this.dy = 0;
            this.scrollCount = 0;
        }
    };

    //現在のX座標を返す
    //１マス５０ピクセル分なので、座標は「this.x * 50」となり、カウンタ１で１０ピクセル分の移動になるため、その分を加算している
    this.getScrollX = function () {
        return this.x * 50 + this.dx * this.scrollCount * 10;//移動中スクロール量を含むx座標
    };

    //現在のY座標を返す
    this.getScrollY = function () {
        return this.y * 50 + this.dy * this.scrollCount * 10;//移動中スクロール量を含むy座標
    };
}

//主人公オブジェクトコンストラクタ
function Player(x, y) {
    this.x = x;//現在の主人公オブジェクトのX座標
    this.y = y;//現在の主人公オブジェクトのY座標
    this.dx = 0;//主人公オブジェクトのX方向移動量
    this.dy = 0;//主人公オブジェクトのY方向移動量
    this.dir = 0;//主人公の上下左右の画像の向き
    this.scrollCount = 0;//５等分された隣りのマスへの移動のうち、何個目に当たるかを数えるカウンタ

    this.update = function () {
        this.doScroll();//スクロール移動。Scrollerオブジェクトのスクロール用カウンタscrollCountを増やす
        if (this.scrollCount > 0) {
            return;//カウンタscrollCountの値が０より大きい場合は、スクロール中なので何もしない
        }

        if (this.x == W - 2 && this.y == H - 2) {//座標(W-2, H-2)＝迷路右下
            clearInterval(timer);//主人公の座標が(W-2, H-2)のときゲームクリア
            game = GAMECLEAR;
            document.getElementById("bgm").pause();
            repaint();
        }

        this.dx = 0;//X方向移動量リセット
        this.dy = 0;//Y方向移動量リセット
        let nx = 0;//仮のX方向移動量
        let ny = 0;//仮のY方向移動量
        switch (code) {
            case "ArrowLeft":
                nx = -1;
                this.dir = 2;
                break;
            case "ArrowUp":
                ny = -1;
                this.dir = 0;
                break;
            case "ArrowRight":
                nx = +1;
                this.dir = 3;
                break;
            case "ArrowDown":
                ny = +1;
                this.dir = 1;
                break;
        }
        if (maze[this.y + ny][this.x + nx] == 0) {//移動先の座標が通路(0)のとき
            this.dx = nx;//X方向移動量を設定
            this.dy = ny;//Y方向移動量を設定
        }
    };

    this.paint = function (gc, x, y, w, h) {
        let img = document.getElementById("hero" + this.dir);
        gc.drawImage(img, x, y, w, h);//座標(x,y)に大きさ(w,h)で主人公描画
    };
}

//敵オブジェクトコンストラクタ
function Alien(x, y) {
    this.x = x;//現在の敵オブジェクトのX座標
    this.y = y;//現在の敵オブジェクトのY座標
    this.dx = 0;//敵オブジェクトのX方向移動量
    this.dy = 0;//敵オブジェクトのY方向移動量
    this.dir = 0;//敵の上下左右の画像の向き
    this.scrollCount = 0;//５等分された隣りのマスへの移動のうち、何個目に当たるかを数えるカウンタ

    this.update = function () {
        this.doScroll();//スクロール移動。Scrollerオブジェクトのスクロール用カウンタscrollCountを増やす

        //主人公との衝突判定（X軸、Y軸の距離の差が40以下⇒衝突）
        let diffX = Math.abs(player.getScrollX() - this.getScrollX()); // X方向差分
        let diffY = Math.abs(player.getScrollY() - this.getScrollY()); // Y方向差分
        if (diffX <= 40 && diffY <= 40) {
            clearInterval(timer);//衝突時処理-->ゲームオーバー
            game = GAMEOVER;
            document.getElementById("bgm").pause();
            repaint();
        }

        //敵の次の移動先を乱数で求める
        let gapx = player.x - this.x;//主人公とのX方向の差分
        let gapy = player.y - this.y;//主人公とのY方向の差分
        switch (random(4)) {
            case 0://X軸方向に近づく
                this.dx = gapx > 0 ? 1 : -1;//gapxの値が正の場合、主人公が敵よりも右にいるので、移動方向dxは１としている
                this.dir = this.dx == -1 ? 2 : 3;
                break;
            case 1://Y軸方向に近づく
                this.dy = gapy > 0 ? 1 : -1;//gapyの値が正の場合、主人公が敵よりも下にいるので、移動方向dyは１としている
                this.dir = this.dy == -1 ? 0 : 1;
                break;
            default://移動しない
                this.dx = 0;
                this.dy = 0;
                break;
        }
    };

    this.paint = function (gc, w, h) {
        let img = document.getElementById("alien" + this.dir);
        gc.drawImage(img, this.getScrollX(), this.getScrollY(), w, h);//敵描画
    };
}

function random(v) {
    return Math.floor(Math.random() * v);//0からvまでの乱数を整数で返す
}

function init() {
    let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
    let maze = document.getElementById("maze");//<canvas>要素への参照を取得
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    maze.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    maze.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = maze.getContext("2d");//描画コンテキスト
    ctx.font = "bold 3rem sans-serif";

    createMaze(W, H);//迷路データ作成
    repaint();
}

function go() {
    //キー用のイベントハンドラを登録
    window.onkeydown = mykeydown;
    window.onkeyup = mykeyup;

    let maze = document.getElementById("maze");//迷路への参照を取得して各種イベントハンドラ登録
    //マウス用のイベントハンドラを登録
    maze.onmousedown = mymousedown;
    maze.onmouseup = mykeyup;
    maze.oncontextmenu = function (e) {
        e.preventDefault();//コンテキストメニューを非表示に(タッチの長押しによってコンテキストメニューが表示される挙動を防止するため)
    };
    //タッチ用のイベントハンドラを登録
    maze.addEventListener("touchstart", mymousedown);
    maze.addEventListener("touchend", mykeyup);

    timer = setInterval(tick, 45);
    document.getElementById("START").style.display = "none";//STARTボタンを非表示に
    document.getElementById("bgm").play();//BGMの再生を開始
}

//メインルーチン
function tick() {
    player.update();
    aliens.forEach((a) => a.update());
    repaint();
}

//棒倒し法(格子状に柱を立て、その柱を４方向のどちらかに倒していく事で迷路を作成する方法)：幅:w、高さ:hの迷路生成
function createMaze(w, h) {
    for (let y = 0; y < h; y++) {
        maze[y] = [];
        for (let x = 0; x < w; x++) {
            //周囲(最も外側のマス目)は壁(1)、それ以外は通路(0)で初期化
            maze[y][x] = x == 0 || x == w - 1 || y == 0 || y == h - 1 ? 1 : 0;
        }
    }

    //左上の隅を(0,0)としたとき、縦・横とも座標が偶数になるマス目に基準となる壁を配置する
    for (let y = 2; y < h - 2; y += 2) {
        for (let x = 2; x < w - 2; x += 2) {
            maze[y][x] = 1;//１つおきに柱を立てる

            //最上段の柱(y=2)は上下左右４方向のどれか、それ以外は下左右の３方向のどれかに柱を倒し、柱を倒した場所は壁にする(最上段以外が上の方向に壁を置かないのは閉路ができるのを防ぐため)
            let dir = random(y == 2 ? 4 : 3);
            let px = x;//今のx座標
            let py = y;//今のy座標
            switch (dir) {
                case 0:
                    py++;//下に倒す
                    break;
                case 1:
                    px--;//左に倒す
                    break;
                case 2:
                    px++;//右に倒す
                    break;
                case 3:
                    py--;//上に倒す
                    break;
            }
            maze[py][px] = 1;//倒れた場所も柱にする
        }
    }
}

//x, yの場所に半径rの円を色colorで描画
function drawCircle(x, y, r, color) {
    ctx.fillStyle = color;
    ctx.beginPath();
    ctx.arc(x, y, r, 0, Math.PI * 2);
    ctx.fill();
}

//描画
function repaint() {
    maze.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    maze.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる

    //背景クリア
    ctx.fillStyle = "black";
    ctx.fillRect(0, 0, 900, 600);

    //クリップ領域設定
    //clip()を実行した後、beginPath()を実行してもマスクは解除されない。
    //clip()を実行する前に、save()で初期状態を保存しておき、マスクが必要なくなったらrestore()で初期状態を復元する
    ctx.save();//save()メソッドは、コンテキストの現在の設定を保存する
    ctx.beginPath();
    ctx.arc(300, 300, 300, 0, Math.PI * 2);
    //clip()メソッドは、パスに登録された図形の外側にマスクをかけて、図形内だけ描画できるようにする
    //clip()メソッドにより、ctx.restore()が呼び出されるまで、描画処理が作成された円の中だけに限定される
    ctx.clip();

    //画面中央の迷路描画
    //主人公が常に中心に描画され、迷路を主人公と反対側に動かすことで、スクロール効果を演出
    ctx.fillStyle = "brown";
    //ctx.translate()で迷路と主人公の座標系を移動する処理を行う
    ctx.translate(6 * 50, 6 * 50);
    ctx.translate(-1 * player.getScrollX(), -1 * player.getScrollY());
    for (let x = 0; x < W; x++) {
        for (let y = 0; y < H; y++) {
            if (maze[y][x] == 1) {
                ctx.fillRect(x * 50, y * 50, 50, 50);//壁の画像描画
            }
        }
    }
    aliens.forEach((a) => a.paint(ctx, 50, 50));//敵を描画
    ctx.restore();//restore()メソッドは、save()メソッドで保存された設定を復元する

    //画面右にある小さな地図描画
    ctx.fillStyle = "#eeeeee";
    ctx.fillRect(650, 0, 250, 600);

    ctx.save();
    ctx.translate(670, 300);
    ctx.fillStyle = "brown";
    for (let x = 0; x < W; x++) {
        for (let y = 0; y < H; y++) {
            if (maze[y][x] == 1) {
                ctx.fillRect(x * 7, y * 7, 7, 7);
            }
        }
    }
    drawCircle(player.x * 7 + 3, player.y * 7 + 3, 3, "red");//自分を赤で
    aliens.forEach((a) => {
        drawCircle(a.x * 7 + 3, a.y * 7 + 3, 3, "purple");//敵を紫で
    });

    ctx.restore();

    //上下左右移動コントローラで、押下状態の〇を描画
    ctx.drawImage(arrows, 670, 70, 200, 200);
    let ax = -100;
    let ay = -100;
    switch (code) {
        case "ArrowRight":
            ax = 830;
            ay = 170;
            break;
        case "ArrowDown":
            ax = 770;
            ay = 230;
            break;
        case "ArrowLeft":
            ax = 710;
            ay = 170;
            break;
        case "ArrowUp":
            ax = 770;
            ay = 120;
            break;
    }
    drawCircle(ax, ay, 30, "yellow");

    //主人公描画とメッセージ
    player.paint(ctx, 300, 300, 50, 50);
    ctx.fillStyle = "yellow";
    if (game == GAMEOVER) {
        ctx.fillText("GAME OVER", 150, 200);
    } else if (game == GAMECLEAR) {
        ctx.fillText("GAME CLEAR", 150, 200);
    }
}

//キー押下時のイベントハンドラ
function mykeydown(e) {
    code = e.code;
}
//キーを離した時のイベントハンドラ
function mykeyup(e) {
    code = "";
}
//マウスクリックやタッチ時のイベントハンドラ
function mymousedown(e) {
    //マウスがクリックされている場合には、e.offsetXに数値が格納されているので、!isNaN(e.offsetX)=trueになり、e.offsetXの値がmouseXに設定される
    //e.offsetXに数値以外が格納されている場合には、e.touches[0].clientXでタッチ座標をmouseXに設定する
    let mouseX = !isNaN(e.offsetX) ? e.offsetX : e.touches[0].clientX;//X座標取得
    let mouseY = !isNaN(e.offsetY) ? e.offsetY : e.touches[0].clientY;//Y座標取得
    //どのボタンの上でクリック又はタッチされたのかを判定する処理
    if (670 < mouseX && mouseX < 870 && 70 < mouseY && mouseY < 270) {//座標値が上下左右ボタンの領域に含まれていれば
        mouseX -= 770;
        mouseY -= 170;
        if (Math.abs(mouseX) > Math.abs(mouseY)) {//左右ボタンを判定
            code = mouseX < 0 ? "ArrowLeft" : "ArrowRight";
        } else {
            code = mouseY < 0 ? "ArrowUp" : "ArrowDown";
        }
    }
}
