"use strict";//より厳密なエラーチェックを行える

//グローバル変数
//重みづけ(どこに優先して石を置くか)計算用の配列
let WeightData = [
    [30, -12, 0, -1, -1, 0, -12, 30],
    [-12, -15, -3, -3, -3, -3, -15, -12],
    [0, -3, 0, -1, -1, 0, -3, 0],
    [-1, -3, -1, -1, -1, -1, -3, -1],
    [-1, -3, -1, -1, -1, -1, -3, -1],
    [0, -3, 0, -1, -1, 0, -3, 0],
    [-12, -15, -3, -3, -3, -3, -15, -12],
    [30, -12, 0, -1, -1, 0, -12, 30]
];
let BLACK = 1;//プレイヤー(黒石)
let WHITE = 2;//PC(白石)
let data = [];//盤面における石の配置状態を保持する配列（0:なし、1:黒、2:白）
let myTurn = false;//自分の番か否か

//初期化関数
function init() {
    let b = document.getElementById("board");
    for (let i = 0; i < 8; i++) {
        let tr = document.createElement("tr");//行を作成
        data[i] = [0, 0, 0, 0, 0, 0, 0, 0];//i行目のデータ設定
        for (let j = 0; j < 8; j++) {
            let td = document.createElement("td");//列を作成
            td.className = "cell";
            td.id = "cell" + i + j;//<td>要素のidを設定
            td.onclick = clicked;//マス目をクリックした時のイベントハンドラを登録
            tr.appendChild(td);
        }
        b.appendChild(tr);
    }
    //最初に4つの石を配置
    put(3, 3, BLACK);
    put(4, 4, BLACK);
    put(3, 4, WHITE);
    put(4, 3, WHITE);
    update();
}

function update() {
    let numWhite = 0;//盤に置かれた白の数
    let numBlack = 0;//盤に置かれた黒の数
    for (let x = 0; x < 8; x++) {
        for (let y = 0; y < 8; y++) {
            if (data[x][y] == WHITE) {
                numWhite++;
            }
            if (data[x][y] == BLACK) {
                numBlack++;
            }
        }
    }
    document.getElementById("numBlack").textContent = numBlack;
    document.getElementById("numWhite").textContent = numWhite;

    let blackFlip = canFlip(BLACK);//黒石をひっくり返せるか否か
    let whiteFlip = canFlip(WHITE);//白石をひっくり返せるか否か

    if (numWhite + numBlack == 64 || (!blackFlip && !whiteFlip)) {
        if (numWhite > numBlack) {
            document.getElementById("message").textContent = "白の勝ち！";
        } else if (numBlack > numWhite) {
            document.getElementById("message").textContent = "黒の勝ち！";
        } else {
            document.getElementById("message").textContent = "引き分け";
        }
        return
    }

    if (!blackFlip) {
        showMessage("黒スキップ");
        myTurn = false;
    } else if (!whiteFlip) {
        showMessage("白スキップ");
        myTurn = true;
    } else {
        myTurn = !myTurn;//順番を交代する
    }
    if (!myTurn) {
        setTimeout(think, 1000);//PCの順番の場合には、1秒後に処理を実行する様にする
    }
}

//メッセージ(str)を2秒間表示
function showMessage(str) {
    document.getElementById("message").textContent = str;
    setTimeout(() => {
        document.getElementById("message").textContent = "";
    }, 2000);
}

//盤上のセルクリック時のコールバック関数(<td>要素のイベントハンドラ)
function clicked(e) {
    if (!myTurn) {
        return;//PC考え中
    }
    let id = e.target.id;//クリックされた<td>要素のidを取得
    let i = parseInt(id.charAt(4));//行番号
    let j = parseInt(id.charAt(5));//列番号

    let flipped = getFlipCells(i, j, BLACK);
    if (flipped.length > 0) {//ひっくり返せる石があった場合
        for (let k = 0; k < flipped.length; k++) {
            put(flipped[k][0], flipped[k][1], BLACK);//返せる石を反転させてその場所に黒石を置く
        }
        put(i, j, BLACK);//クリックしたセルに黒石を置く
        update();
    }
}

//(i,j)にcolor色の石を置く
function put(i, j, color) {
    let c = document.getElementById("cell" + i + j);
    c.textContent = "●";
    c.className = "cell " + (color == BLACK ? "black" : "white");//クラスセレクタをcell blackの様に2つ設定するために、cellの後に半角スペースを入れる
    data[i][j] = color;
}

//コンピュータの思考関数
function think() {
    let highScore = -1000;
    let px = -1;
    let py = -1;
    //１．64マス中で白を置ける場所を探し、そこに白石置いたときの状態を再現する
    for (let x = 0; x < 8; x++) {
        for (let y = 0; y < 8; y++) {
            let tmpData = copyData();//現在の盤面をコピーしたものを「仮にこの場所に白を置けばどうなるか」を再現するための仮データtmpDataに代入
            let flipped = getFlipCells(x, y, WHITE);//(x,y)座標に白を置いたときに黒から白に反転する石の配列を求める
            if (flipped.length > 0) {//黒から白に反転する石がある場合
                for (let i = 0; i < flipped.length; i++) {
                    let p = flipped[i][0];
                    let q = flipped[i][1];
                    tmpData[p][q] = WHITE;
                    tmpData[x][y] = WHITE;
                }
                //２．その状態において優先順位の合計値(白石のある場所の点の合計)を求める
                let score = calcWeightData(tmpData);
                if (score > highScore) {//３．優先順位の合計値が最大になった場所に石を置く
                    highScore = score;
                    (px = x), (py = y);//現在の(x,y)座標をpxとpyに保存する
                }
            }
        }
    }
    if (px >= 0 && py >= 0) {//pxとpyは、この関数の冒頭で共に-1で初期化しているので、両方とも0以上ということはどこかに石を置けたことを意味する
        let flipped = getFlipCells(px, py, WHITE);
        if (flipped.length > 0) {
            for (let k = 0; k < flipped.length; k++) {
                put(flipped[k][0], flipped[k][1], WHITE);
            }
        }
        put(px, py, WHITE);
    }
    update();
}

//引数で与えられた盤面データの優先順位の合計値を求めて返す関数
function calcWeightData(tmpData) {
    let score = 0;
    for (let x = 0; x < 8; x++) {
        for (let y = 0; y < 8; y++) {
            if (tmpData[x][y] == WHITE) {
                score += WeightData[x][y];
            }
        }
    }
    return score;
}

//石テーブルデータをコピー
function copyData() {
    let tmpData = [];
    for (let x = 0; x < 8; x++) {
        tmpData[x] = [];
        for (let y = 0; y < 8; y++) {
            tmpData[x][y] = data[x][y];//現在の盤面の状態を仮データにコピー
        }
    }
    return tmpData;
}

//盤面に引数の色の石を置けるか否か(挟めるかどうか)？
function canFlip(color) {
    for (let x = 0; x < 8; x++) {
        for (let y = 0; y < 8; y++) {
            let flipped = getFlipCells(x, y, color);
            if (flipped.length > 0) {//(x,y)座標に石を置いたときにひっくり返せる数
                return true;
            }
        }
    }
    return false;
}

//(i,j)に石を置いたときに石を挟めるか？
function getFlipCells(i, j, color) {
    if (data[i][j] == BLACK || data[i][j] == WHITE) {
        return [];//すでに石があるときは空の配列を返す
    }
    //石を置いたときに相手の石を挟める場合の処理
    //相手の石を挟めるかどうかを石を置こうとしている場所の、左上、上、右上、左、右、左下、下、右下と順番に調査
    let dirs = [//上記の８方向を配列に格納
        [-1, -1],
        [0, -1],
        [1, -1],
        [-1, 0],
        [1, 0],
        [-1, 1],
        [0, 1],
        [1, 1]
    ];
    let result = [];
    for (let p = 0; p < dirs.length; p++) {//各方向ごとに相手の石を挟めるか否かを判定
        let flipped = getFlipCellsOneDir(i, j, dirs[p][0], dirs[p][1], color);
        result = result.concat(flipped);
    }
    return result;
}

//(i,j)に石を置いたときに、(dx,dy)方向で石を挟めるか？
function getFlipCellsOneDir(i, j, dx, dy, color) {
    //(i,j)座標の１つ先の座標(隣の座標)
    let x = i + dx;
    let y = j + dy;
    let flipped = [];//挟まれた石の配列

    if (x < 0 || y < 0 || x > 7 || y > 7 || data[x][y] == color || data[x][y] == 0) {
        return [];//隣が盤外、同色、石が無い場合には空の配列を返す（挟めない）
    }
    flipped.push([x, y]);//別の色の石が(dx,dy)方向に隣接している場合、その石の座標を仮に配列flippedに格納する

    while (true) {
        //(i,j)座標から２つ先の座標以降について
        x += dx;
        y += dy;
        if (x < 0 || y < 0 || x > 7 || y > 7 || data[x][y] == 0) {
            return [];//盤外、石が無い場合には空の配列を返す（挟めない）
        }
        if (data[x][y] == color) {//その座標に同じ色の石があった場合
            return flipped;//挟めた！
        } else {//別の色の石が連続している場合
            flipped.push([x, y]);//while文の実行が継続される
        }
    }
}
