"use strict";

//棒人間クラス
class Jumper {
    constructor() {
        this.y = 200;//座標
        this.sy = 0;//加速度(1フレームでの上下方向移動量)
        this.index = 0;//アニメーション画像のインデックス番号
        this.images = [];//画像の配列
        this.landed = true;//着地してるか否か
        //アニメーション用の画像を８つ読み込み、imagesプロパティに格納
        for (let i = 0; i < 8; i++) {
            this.images.push(document.getElementById("stick" + i));
        }
    }
    //ジャンプ時に呼び出されるメソッド
    jump(power) {
        //着地時(landedプロパティがtrue)にはpowerに応じてジャンプ
        if (this.landed) {
            this.y -= 10;//主人公が箱と衝突しないようにY座標を10減らす
            this.sy = -power;//上方向の初速
        }
    }
    //主人公(忍者)の描画と衝突処理を行うメソッド
    paint() {
        //ジャンプ中 : 画像のコマを進め、加速度を加えて、上下方向に移動
        if (!this.landed) {
            this.index = (this.index + 1) % this.images.length;
            this.sy += 0.3;//重力加速度がかかるようにY軸方向の速度を0.3増やす
            this.y += this.sy;
        }
        ctx.drawImage(this.images[this.index], 100, this.y);//主人公(忍者)の描画
        this.landed = false;
        boxes.forEach((b) => {
            let foot = this.y + 150;//足のY座標
            let right = b.x + b.w;//箱の右端座標
            //主人公(忍者)が箱に衝突した場合
            if (b.x < 150 && right > 150 && b.y <= foot && b.y + 40 >= foot) {
                this.landed = true;
                this.y = b.y - 150;//箱の上に主人公(忍者)が来るように
            }
        });
    }
}

//箱クラス
class Box {
    constructor(x, y, w) {
        this.x = x;//箱の左上のX座標
        this.y = y;//箱の左上のY座標
        this.w = w;//幅
    }
    paint() {
        ctx.fillStyle = 'rgba(50, 0, 0, 0.5)';
        ctx.fillRect(this.x, this.y, this.w, 100);//矩形を塗りつぶす
    }
}

let ctx;//描画コンテキスト
let jumper;//主人公(忍者)
let power = 0;//ジャンプの力
let timerId = NaN;//タイマー
let isMouseDown = false;//マウス押下中か否か
let boxes = [new Box(0, 350, 300)];//箱のリスト(最初の１つは設定)
let back;//背景画像
let code = 0;//現在どのキーが押されているか
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let field = document.getElementById("field");//<canvas>要素への参照を取得

onload = () => {
    jumper = new Jumper(100, 200);//主人公(忍者)オブジェクト作成
    for (let x = 400; x < 3000;) {
        //箱をリストに追加(X軸で3000まで)
        let w = Math.random() * 200 + 100;
        let y = Math.random() * 300 + 300;
        boxes.push(new Box(x, y, w));
        x += w + Math.random() * 200 + 100;
    }

    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    field.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    field.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = field.getContext("2d");//描画コンテキスト
    ctx.font = "1rem 'sans-serif'";
    ctx.strokeStyle = "#000000";
    ctx.lineWidth = 5;
    back = document.getElementById("back");

    //イベントハンドラの登録(マウス操作)
    window.onpointerdown = () => {
        isMouseDown = true;//マウス押下中
    };
    window.onpointerup = () => {
        isMouseDown = false;//マウスを離した→ジャンプ
        jumper.jump(power / 20);
    };
    //イベントハンドラの登録(矢印キー操作)
    document.addEventListener('keydown', keydown_arrowup);
    function keydown_arrowup(e) {
        if (e.code == "ArrowUp") {
            isMouseDown = true;//上矢印押下中
        }
    }
    document.addEventListener('keyup', keyup_arrowup);
    function keyup_arrowup(e) {
        if (e.code == "ArrowUp") {
            isMouseDown = false;//上矢印を離した→ジャンプ
            jumper.jump(power / 20);
        }
    }

    timerId = setInterval(tick, 50);//タイマーの開始
};

//メインループ
function tick() {
    power = isMouseDown ? Math.min(power + 10, 600) : 0;//マウス押下→パワー増加
    ctx.fillStyle = "#ffffff";
    ctx.fillRect(0, 0, field.width, field.height);
    ctx.drawImage(back, 0, 0, field.width, field.height);
    ctx.fillStyle = "#000000";
    ctx.fillRect(0, 0, power, 15);//powerゲージを描画
    ctx.fillText("パワー：" + power, 610, 15);
    jumper.paint();//主人公(忍者)描画
    boxes.forEach((b) => {
        if (!jumper.landed) {
            b.x -= 5;//ジャンプ中は箱を左へ移動
        }
        b.paint();//箱の描画
    });
    if (jumper.y > 500) {//落下時はゲームオーバー
        clearInterval(timerId);
        ctx.font = "5rem 'sans-serif'";
        ctx.fillText("ゲームオーバー", field.width / 3, field.height / 2);
    }
    //配列boxesには、全ての箱オブジェクトが入っており、その個数はboxes.lengthで取得でき、その一番右端のオブジェクトは、boxes[boxes.length - 1]で取得できる
    if (boxes[boxes.length - 1].x < 0) {//すべての箱を飛んだらクリア
        clearInterval(timerId);
        ctx.font = "5rem 'sans-serif'";
        ctx.fillText("ゲームクリア！", field.width / 3, field.height / 2);
    }
}
