"use strict";

let ctx;//描画コンテキスト
let tiles = [];//タイルオブジェクト(ブロック)を格納する二次元配列
let moves = [];//移動中のタイル(ブロック)を保持する配列
let mIndex = 0;//連鎖メッセージのインデックス(＝何連鎖中かを保持)
let mCount = 0;//連鎖メッセージのフェードアウト効果(色の濃さ)を演出するためのカウンタ
let times = [];//残り時間画像を格納する配列
let timer = NaN;//タイマー
let startTime = NaN;//ゲーム開始時刻
let elapsed = 0;//経過時間
let score = 0;//スコア
let bgimage;//背景画像
let sound;//ブロックが消えたときの効果音
let mouseX = null;//マウス押下時のX座標
let mouseY = null;//マウス押下時のY座標
let mouseUpX = null;//マウスリリース時のX座標
let mouseUpY = null;//マウスリリース時のY座標
//連鎖メッセージの配列
let message = [
    "",
    "good",
    "very good",
    "super",
    "wonderful!",
    "great!!",
    "amazing",
    "brilliant!",
    "excellent!!",
];

function rand(v) {
    return Math.floor(Math.random() * v);//0～vまでの乱数（整数）を返す
}

//２重ループを使って、全てのタイルに引数である関数オブジェクトfを適用する。
function iterate(f) {
    for (let x = 0; x < 12; x++) {
        for (let y = 0; y < 12; y++) {
            f(x, y, tiles[x][y]);//(x,y)の座標とタイルを引数に関数f()を実行
        }
    }
}

//タイルオブジェクト
function Tile(x, y) {
    //プロパティ
    this.x = x;//現在のX座標
    this.y = y;//現在のY座標
    this.px = x;//移動先のX座標
    this.py = y;//移動先のY座標
    this.count = 0;//移動量計算用カウンタ
    //描画座標を返すメソッド。countの値によってオフセット(実際の位置からどれだけずらすかという量)を調整している
    this.getX = function () {
        return this.x + ((this.px - this.x) * this.count) / 20;//移動中も考慮したX座標
    };
    this.getY = function () {
        return this.y + ((this.py - this.y) * this.count) / 20;//移動中も考慮したY座標
    };
    //タイルを視覚的に移動させるメソッド
    this.move = function (px, py, color) {
        this.px = px;//移動先のX座標
        this.py = py;//移動先のY座標
        this.color = color;//移動先の色
        this.count = 20;//座標計算用カウンタ（20カウントかけて移動）
        this.moving = true;//移動中フラグ
        moves.push(this);//移動中タイルの格納配列movesに自身を挿入
    };
    this.update = function () {
        if (--this.count <= 0) {
            this.moving = false;//カウンタが0になったら移動中フラグをクリア
        }
    };
}

//ゲーム初期化
function init() {
    //２重ループでタイルオブジェクトの二次元配列を生成
    for (let x = 0; x < 12; x++) {
        tiles[x] = [];
        for (let y = 0; y < 12; y++) {
            tiles[x][y] = new Tile(x, y);
        }
    }

    //３つ連続しないよう初期色の配置(初期化段階で連鎖が起きない様にする)
    iterate((x, y, t) => {//関数オブジェクトの引数は、X座標、Y座標、タイルオブジェクトt
        while (true) {
            let r = rand(5);
            //関数setColorがfalseを返す場合は、色が３つ連続しているので、再度乱数の生成から処理を行う
            if (setColor(x, y, r)) {
                t.color = r;
                break;
            }
        }
    });

    //残り時間の画像を初期化
    for (let i = 0; i <= 15; i++) {
        let t = document.createElement("img");
        t.src = "sources/time" + i + ".png";
        times.push(t);
    }

    //Canvas要素を初期化
    bgimage = document.getElementById("bgimage");//グローバル変数bgimageの前にletやconstを付けるとエラーになる
    let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
    let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    ctx = canvas.getContext("2d");//描画コンテキスト
    ctx.textAlign = "center";

    sound = document.getElementById("sound");
    repaint();
}

//ゲーム実行開始
function go() {
    let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得
    //イベントハンドラ登録
    canvas.onmousedown = mymousedown;//マウス押下時イベントハンドラ登録
    canvas.onmouseup = mymouseup;//マウスリリース時イベントハンドラ登録
    canvas.addEventListener("touchstart", mymousedown);
    canvas.addEventListener("touchmove", mymousemove);
    canvas.addEventListener("touchend", mymouseup);

    startTime = new Date();
    timer = setInterval(tick, 25);

    //タッチで指を動かした時、ブラウザによっては、ゲーム操作ではなく、ゲーム画面を移動する操作と解釈するため記述
    document.body.addEventListener(
        "touchmove",
        (event) => {
            event.preventDefault();//画面タッチによるコンテキストメニューを非表示に
        },
        false
    );
    document.getElementById("START").style.display = "none";//STARTボタンを非表示に
    document.getElementById("bgm").play();//BGM再生開始
}

//メインループ
function tick() {
    mCount = Math.max(0, mCount - 1);// メッセージフェードアウト用にmCountを１減らす(０以上の範囲で)
    if (mCount == 0) {
        mIndex = 0;// mCountが０のとき(連鎖メッセージが消えたとき)、連鎖メッセージのインデックスを０にリセットして連鎖をクリア
    }

    if (moves.length > 0) {
        //移動中のタイル(moves)がある場合に、それらタイルを移動
        for (let i = 0; i < moves.length; i++) {
            moves[i].update();
        }
        //移動が完了していないオブジェクトのみ抽出
        //タイルのcountが0でない(＝まだ移動中のタイル)が配列としてmovesに戻される
        moves = moves.filter((t) => {
            return t.count != 0;
        });
        //全ての移動中のタイル(moves)が移動完了
        if (moves.length == 0) {
            let s = removeTile();//タイル消去(removeTile()関数は、消去されたタイルの個数を戻す）
            if (s > 0) {//連鎖があった場合
                mIndex = Math.min(message.length - 1, mIndex + 1);//次の連鎖メッセージへ
                mCount = 50;//メッセージ透明度カウンタを初期化
                score += s * 10 + mIndex * s * 100;//連鎖に応じてスコア加算
                sound.pause();
                sound.currentTime = 0;
                sound.play();//連鎖用効果音再生
            }
            fall();//ブロック落下
        }
    }

    elapsed = (new Date().getTime() - startTime.getTime()) / 1000;//経過時間
    if (elapsed > 69) {
        clearInterval(timer);//70秒でゲーム終了(BGMの再生時間が69秒のため)
        timer = NaN;
    }
    repaint();
}

/*ビット演算子(0と1で表される２進数の各ビット演算に使われる演算子)
ビット演算の論理積は「&」で表し、各ビットがどちらも「1」の場合のみ「1」になり、どちらかが「0」の場合「0」になる。
ビット演算の論理和は「|」で表し、各ビットのどちらかが「1」であれば「1」になり、どちらも「0」の場合のみ「0」になる。
ビット演算の排他的論理和は「^」で表し、各ビットが「同じ値」であれば「0」になり、異なる値であれば「1」になる。
ビット演算の反転は「~」で表し、各ビットの値を反転させます。「0」ならば「1」に、「1」ならば「0」になる。*/
//初期化時に色を設定する関数。仮に(x,y)の座標に色cを設定した時、上下左右に３つ以上同じ色が隣り合わない場合にはtrueを、そうでない場合はfalseを返す。端に近い場合は連続するタイルがないので、４つのif文を使って上下左右４方向を調べている
function setColor(x, y, c) {
    let flag = true;
    //左方向に3つ連鎖していないか
    if (x > 1) {
        let c0 = tiles[x - 2][y].color;
        let c1 = tiles[x - 1][y].color;
        //引数の色がc、隣の色がc1、さらに隣の色がc0。この３つの色が同じ時、「(c0 == c1 && c1 == c)」がtrueになる。&は、両方がtrueの時のみ結果がtrueになるので、３つの色が同じだと「flag = flag & false」となり、flagもfalseとなる
        flag &= !(c0 == c1 && c1 == c);
    }
    //右方向に3つ連鎖していないか
    if (x < 8) {
        let c0 = tiles[x + 2][y].color;
        let c1 = tiles[x + 1][y].color;
        flag &= !(c0 == c1 && c1 == c);
    }
    //上方向に3つ連鎖していないか
    if (y > 1) {
        let c0 = tiles[x][y - 2].color;
        let c1 = tiles[x][y - 1].color;
        flag &= !(c0 == c1 && c1 == c);
    }
    //下方向に3つ連鎖していないか
    if (y < 8) {
        let c0 = tiles[x][y + 2].color;
        let c1 = tiles[x][y + 1].color;
        flag &= !(c0 == c1 && c1 == c);
    }
    //全ての方向で３つ以上色が揃っていない場合のみtrueが返り、一つの方向でも３つ以上色が揃っている場合にはfalseが返される
    return flag;
}

function mymousedown(e) {
    mouseX = !isNaN(e.offsetX) ? e.offsetX : e.touches[0].clientX;//マウス押下＆タッチ操作時のX座標
    mouseY = !isNaN(e.offsetY) ? e.offsetY : e.touches[0].clientY;//マウス押下＆タッチ操作時のY座標
}

//特定のブラウザでタッチ操作が行われた場合に、mymouseup(e)関数ではリリース時の座標値を取得できない事があるため、そのケースに備えて最後に移動した際の座標を取得し、保存する
function mymousemove(e) {
    mouseUpX = !isNaN(e.offsetX) ? e.offsetX : e.touches[0].clientX;//マウスリリースX座標
    mouseUpY = !isNaN(e.offsetY) ? e.offsetY : e.touches[0].clientY;//マウスリリースY座標
}

//マウスが離された時のコールバック関数。タイル領域の左上座標は(34,36)、タイル１つ分の幅と高さが44px
function mymouseup(e) {
    let sx = Math.floor((mouseX - 34 - canvas.width / 4) / 44);//移動元(マウスをクリックした)タイルのX番号
    let sy = Math.floor((mouseY - 36) / 44);//移動元(マウスをクリックした)タイルのY番号
    let nx = sx;//移動先タイルのX番号
    let ny = sy;//移動先タイルのY番号
    let mx = !isNaN(e.offsetX) ? e.offsetX : mouseUpX;//リリース時のX座標
    let my = !isNaN(e.offsetY) ? e.offsetY : mouseUpY;//リリース時のY座標

    //押下時(mouseX,mouseY)からリリース時(mx,my)のX軸・Y軸方向の移動量を調べて、その内大きい方に移動
    if (Math.abs(mx - mouseX) > Math.abs(my - mouseY)) {
        nx += mx - mouseX > 0 ? 1 : -1;//X軸方向に移動
    } else {
        ny += my - mouseY > 0 ? 1 : -1;//Y軸方向に移動
    }

    if (nx > 11 || ny > 11 || nx < 0 || ny < 0) {
        return;//移動先が範囲外のときは何もしない
    }
    if (tiles[sx][sy].moving || tiles[nx][ny].moving) {
        return;//対象となるタイルが移動中のときは何もしない
    }

    //上記２つ以外の場合、(sx,sy)と(nx,ny)のタイルの色を入れ替え
    let c = tiles[sx][sy].color;
    tiles[sx][sy].move(nx, ny, tiles[nx][ny].color);
    tiles[nx][ny].move(sx, sy, c);
    repaint();
}

//横方向、縦方向に３つ以上同じ色が並んだ時に、そのタイルオブジェクトのremoveプロパティにtrueを設定する
function removeTile() {
    //横方向に3つ以上連続するタイルにremoveフラグをセット
    for (let y = 0; y < 12; y++) {
        let c0 = tiles[0][y].color;//0列目(左端)の色
        let count = 1;//同色が連続する数
        for (let x = 1; x < 12; x++) {//左端(x=0)の隣のタイルから調べている
            let c1 = tiles[x][y].color;//右方向に移動しながらタイルの色を取得
            if (c0 != c1) {
                //色が異なる場合はその色から数えなおし
                c0 = c1;
                count = 1;
            } else {
                if (++count >= 3) {
                    //連続する個数が3を超えた場合、そのタイルにremoveフラグをセット
                    tiles[x - 2][y].remove = true;
                    tiles[x - 1][y].remove = true;
                    tiles[x - 0][y].remove = true;
                }
            }
        }
    }

    //縦方向に3つ以上連続するタイルにremoveフラグをセット
    for (let x = 0; x < 12; x++) {
        let c0 = tiles[x][0].color;//0行目の色
        let count = 1;
        for (let y = 1; y < 12; y++) {
            let c1 = tiles[x][y].color;
            if (c0 != c1) {
                c0 = c1;
                count = 1;
            } else {
                if (++count >= 3) {
                    //連続する個数が3を超えた場合、そのタイルにremoveフラグをセット
                    tiles[x][y - 2].remove = true;
                    tiles[x][y - 1].remove = true;
                    tiles[x][y - 0].remove = true;
                }
            }
        }
    }
    //removeプロパティがtrueに設定されたタイル(削除したタイル)の個数を戻り値として返す（スコア計算用）
    let removed = 0;
    iterate((x, y, t) => {
        if (t.remove) {
            removed++;
        }
    });
    return removed;
}

//落下処理
function fall() {
    for (let x = 0; x < 12; x++) {
        //横(X軸)方向は左列から右列へ順番に処理を行う
        for (let y = 11, sp = 11; y >= 0; y--, sp--) {
            //縦(Y軸)方向は下から上へ(yは現在のタイルを示す値。spは落下してくるタイルを示す値で、削除されたタイルをスキップするカウンタ)
            while (sp >= 0) {
                if (tiles[x][sp].remove) {
                    sp--;//removeプロパティが追加されているタイルがある場合、spはそれをスキップするためにさらにデクリメントする
                } else {
                    break;
                }
            }
            //yとspの値が異なる（削除されたタイルがあった） → タイルを移動する
            if (y != sp) {
                //spが0未満の時は落ちてくるタイルがないので乱数で色を設定
                let c = sp >= 0 ? tiles[x][sp].color : rand(5);
                tiles[x][y].move(x, sp, c);//移動元tiles[x][sp]、移動先tiles[x][y]
            }
        }
    }
    iterate((x, y, t) => {
        t.remove = false;//すべてのタイルでremoveフラグをリセット
    });
}

function repaint() {
    canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる

    ctx.drawImage(bgimage, canvas.width / 4, 0);

    //タイル描画
    let images = [block0, block1, block2, block3, block4];
    iterate((x, y, t) => {
        if (!t.remove) {
            ctx.drawImage(images[t.color], t.getX() * 44 + 34 + canvas.width / 4, t.getY() * 44 + 36, 42, 42);
        }
    });

    //連鎖メッセージ描画
    ctx.font = "bold 5rem sans-serif";
    ctx.fillStyle = "rgba(255,255,255," + mCount / 50 + ")";//連鎖メッセージの濃度を調整
    ctx.fillText(message[mIndex], canvas.width / 3, canvas.height / 9);
    ctx.fillStyle = "white";

    if (isNaN(timer)) {
        ctx.fillText("FINISH", canvas.width / 3, canvas.height / 2);
    }

    //スコア
    ctx.fillStyle = "rgba(220,133,30,50)";
    ctx.font = "bold 3rem sans-serif";
    //slice()メソッドの引数にマイナスの値を指定した場合、文字列の末尾から切り取って返される
    ctx.fillText(("0000000" + score).slice(-7), canvas.width / 1.6, canvas.height / 4);

    //残り時間(配列に格納した画像で表現)
    let index = Math.min(15, Math.floor(elapsed / (69 / 15)));
    ctx.drawImage(times[index], canvas.width / 1.5, canvas.height / 2);
}
