"use strict";

let ctx, engine, timer, offset = 0, catcher, score = 0, isMouseDown = false;
let wrap = document.getElementById("wrap");//<canvas>要素の親要素への参照を取得
let canvas = document.getElementById("canvas");//<canvas>要素への参照を取得

let walls = [
    [-100 + canvas.width * 1.69, -100, 100, 1000],
    [-100 + canvas.width * 1.69, -100, 800, 100],
    [500 + canvas.width * 1.69, -100, 100, 1000]
];

let lines = [
    [150 + canvas.width * 1.69, -50, -50 + canvas.width * 1.69, 150],
    [350 + canvas.width * 1.69, -50, 550 + canvas.width * 1.69, 150],
    [450 + canvas.width * 1.69, 200, 450 + canvas.width * 1.69, 1000]
];

function init() {
    //getBoundingClientRect()メソッドは、borderは含み、marginは含まない要素の寸法と、そのビューポートに対する相対位置に関する情報をFloat値で返す
    canvas.width = wrap.getBoundingClientRect().width;//<canvas>要素の幅を親要素の幅と合わせる
    canvas.height = wrap.getBoundingClientRect().height;//<canvas>要素の高さを親要素の高さと合わせる
    //エンジン初期化 & イベントハンドラ設定
    engine = new Engine(-100, -100, canvas.width, canvas.height, 0, 9.8);
    canvas.onmousedown = mymousedown;
    canvas.onmouseup = mymouseup;
    canvas.addEventListener('touchstart', mymousedown);
    canvas.addEventListener('touchend', mymouseup);
    //右クリックメニューを無効化
    canvas.oncontextmenu = function (e) { e.preventDefault(); };

    //壁を作成してエンジンに追加
    walls.forEach((w) => {
        let r = new RectangleEntity(w[0], w[1], w[2], w[3]);
        r.color = "gray";
        engine.entities.push(r);
    });

    lines.forEach((w) => {
        let r = new LineEntity(w[0], w[1], w[2], w[3], 0.8);
        r.color = "gray";
        engine.entities.push(r);
    });

    //釘
    for (let i = 0; i < 9; i++) {
        for (let j = 0; j < 8 + i % 2; j++) {
            let x = (j * 50 + 50) - 25 * (i % 2);
            let r = new CircleEntity(x + canvas.width / 3, i * 50 + 100, 5, BodyStatic, 1);
            r.color = "blue";
            engine.entities.push(r);
        }
    }

    catcher = new RectangleEntity(canvas.width / 3, 550, 150, 25);
    catcher.color = "gold";
    catcher.sign = 1;

    engine.entities.push(catcher);

    // その他(Canvas, Timer)の初期化
    ctx = canvas.getContext("2d");
    ctx.font = "2rem Arial";
    ctx.strokeStyle = "blue";
    timer = setInterval(tick, 50);
}

function tick() {
    if (isMouseDown) {
        offset = Math.min(offset + 5, 200);
    }
    catcher.sign *= (catcher.x > 300 + canvas.width / 3 || catcher.x < canvas.width / 3) ? -1 : 1;
    catcher.x = catcher.x + 5 * catcher.sign;

    engine.step(0.01);//物理エンジンの時刻を進める
    repaint();//再描画
}

function mymousedown(e) {
    isMouseDown = true;
}

function mymouseup(e) {
    isMouseDown = false;
    let r = new CircleEntity(475 + canvas.width / 3, 400, 10, BodyDynamic);
    r.color = "yellow";
    r.velocity.y = -offset / 5;
    //onhit()メソッド(自身が他のオブジェクトと衝突した時に呼び出される)
    r.onhit = function (me, peer) {
        if (peer == catcher) {
            engine.entities = engine.entities.filter(function (e) {
                return e != me;
            });
            score++;
        }
    }

    offset = 0;
    engine.entities.push(r);
}

function repaint() {
    //背景クリア
    ctx.fillStyle = "#006600";
    ctx.fillRect(canvas.width / 3, 0, 500, canvas.height);

    //ボール・壁の描画
    for (let i = 0; i < engine.entities.length; i++) {
        let e = engine.entities[i];
        ctx.fillStyle = e.color;
        switch (e.shape) {
            case ShapeCircle://円の場合
                ctx.beginPath();
                ctx.arc(e.x, e.y, e.radius, 0, Math.PI * 2);
                ctx.closePath();
                ctx.fill();
                break;
            case ShapeRectangle://矩形の場合
                ctx.fillRect(e.x, e.y, e.w, e.h);
                break;
            case ShapeLine://線分の場合
                ctx.beginPath();
                ctx.lineWidth = 10;
                ctx.moveTo(e.x0, e.y0);
                ctx.lineTo(e.x1, e.y1);
                ctx.stroke();
                break;
        }
    }

    ctx.fillText("score:" + score, + canvas.width / 2.2, 30);
    ctx.fillStyle = "yellow";
    ctx.beginPath();
    ctx.arc(475 + canvas.width / 3, 390 + offset, 10, 0, Math.PI * 2);
    ctx.closePath();
    ctx.fill();

    ctx.fillStyle = "gray";
    ctx.fillRect(455 + canvas.width / 3, 400 + offset, 40, 200);

    ctx.clearRect(0, 0, canvas.width / 3, canvas.height);
    ctx.clearRect(500 + canvas.width / 3, 0, canvas.width, canvas.height);
}
